@@ -29,7 +29,7 @@
 
 #include <linux/module.h>
 #include <linux/moduleparam.h>
-#include <linux/config.h>
+#include <linux/autoconf.h>
 #include <linux/init.h>
 
 #include <linux/version.h>
@@ -59,6 +59,10 @@ typedef void irqreturn_t;
 #define IRQ_HANDLED
 #endif
 
+#ifndef SA_SHIRQ
+#define SA_SHIRQ IRQF_SHARED
+#endif
+
 #if ( LINUX_VERSION_CODE < KERNEL_VERSION(2,6,9) )
 #define __iomem
 #endif
@@ -202,9 +206,12 @@ do { if (ipw_debug_level & (level)) \
 #define IPW_DEBUG_POWER(f, a...)  IPW_DEBUG(IPW_DL_POWER, f, ## a)
 #define IPW_DEBUG_11H(f, a...)  IPW_DEBUG(IPW_DL_11H, f, ## a)
 
-/* Amount of time after sending an association request before we assume
- * the request has failed in ms (default 750ms)*/
-#define ASSOCIATE_TIMEOUT 750
+/* Amount of time after sending an association request to wait before we send
+ * the next request frame. */
+#define IPW_ASSOC_STATE_TIMEOUT 100
+
+/* Number of association sequence request frames to send before we bail */
+#define IPW_MAX_ASSOC_RETRY 7
 
 /* microcode and hardware interface definitions */
 
@@ -1167,6 +1174,7 @@ struct statistics {
 #define ALM_APMG_CLK_DIS                         (0x003008)
 #define ALM_APMG_PS_CTL                          (0x00300c)
 #define ALM_APMG_PCIDEV_STT                      (0x003010)
+#define ALM_APMG_RFKILL				 (0x003014)
 #define ALM_APMG_LARC_INT                        (0x00301c)
 #define ALM_APMG_LARC_INT_MSK                    (0x003020)
 
@@ -1288,21 +1296,13 @@ struct statistics {
 /*********************************************/
 
 /* Authentication  and Association States */
-enum connection_manager_assoc_states {
-	CMAS_INIT = 0,
-	CMAS_TX_AUTH_SEQ_1,
-	CMAS_RX_AUTH_SEQ_2,
-	CMAS_AUTH_SEQ_1_PASS,
-	CMAS_AUTH_SEQ_1_FAIL,
-	CMAS_TX_AUTH_SEQ_3,
-	CMAS_RX_AUTH_SEQ_4,
-	CMAS_AUTH_SEQ_2_PASS,
-	CMAS_AUTH_SEQ_2_FAIL,
-	CMAS_AUTHENTICATED,
-	CMAS_TX_ASSOC,
-	CMAS_RX_ASSOC_RESP,
-	CMAS_ASSOCIATED,
-	CMAS_LAST
+enum ipw_auth_sequence {
+	AUTH_INIT = 0,
+	AUTH_TRANSACTION_1,	/* STA send AUTH request */
+	AUTH_TRANSACTION_2,	/* shared: AP responds */
+	AUTH_TRANSACTION_3,	/* shared: STA sends encrypted text */
+	AUTH_TRANSACTION_4,	/* shared: AP sends success or failure */
+	AUTH_SUCCESS,
 };
 
 #define IPW_POWER_MODE_CAM           0x00	//(always on)
@@ -1532,9 +1532,10 @@ typedef int (*IPW_CALLBACK_FUNC) (struct ipw_priv * priv,
 #define TFD_TX_CMD_SLOTS 64
 #define TFD_CMD_SLOTS 32
 
+#define CMD_VAR_MAGIC 0xA987
 struct ipw_cmd_meta {
+	struct ipw_cmd_meta *source;
 	union {
-		struct ipw_cmd_meta *source;
 		struct sk_buff *skb;
 		IPW_CALLBACK_FUNC callback;
 	} __attribute__ ((packed)) u;
@@ -1546,6 +1547,7 @@ struct ipw_cmd_meta {
 	u8 flags;
 
 	u8 token;
+	u16 magic;
 } __attribute__ ((packed));
 
 struct ipw_cmd_header {
@@ -1966,13 +1968,6 @@ struct ipw_rate_scale_mgr {
 	u32 tx_packets;
 };
 
-/* List of spectrum management request frames */
-struct ipw_spectrum_list {
-	struct list_head list;
-	struct ieee80211_rx_stats stats;
-	struct ieee80211_action action[0];
-};
-
 #define IPW_IBSS_MAC_HASH_SIZE 31
 
 struct ipw_ibss_seq {
@@ -2042,6 +2037,7 @@ struct ipw_link_blink {
 };
 
 struct ipw_frame {
+	int len;
 	union {
 		struct ieee80211_hdr frame;
 		u8 raw[IEEE80211_FRAME_LEN];
@@ -2095,6 +2091,25 @@ struct ipw_rt_hdr {
 } __attribute__ ((packed));
 #endif
 
+/* The following macros are neccessary to retain compatibility
+ * around the workqueue chenges happened in kernels >= 2.6.20:
+ * - INIT_WORK changed to take 2 arguments and let the work function
+ *   get its own data through the container_of macro
+ * - delayed works have been split from normal works to save some
+ *   memory usage in struct work_struct
+ */
+#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 20)
+#define IPW_INIT_WORK(_work, _fun)		INIT_WORK(_work, ((void (*)(void *))_fun), _work)
+#define IPW_INIT_DELAYED_WORK(_work, _fun)	INIT_WORK(_work, ((void (*)(void *))_fun), _work)
+#define IPW_DELAYED_DATA(_ptr, _type, _m)	container_of(_ptr, _type, _m)
+typedef struct work_struct delayed_work_t;
+#else
+#define IPW_INIT_WORK				INIT_WORK
+#define IPW_INIT_DELAYED_WORK			INIT_DELAYED_WORK
+#define IPW_DELAYED_DATA(_ptr, _type, _m)	container_of(_ptr, _type, _m.work)
+typedef struct delayed_work delayed_work_t;
+#endif
+
 struct ipw_priv {
 	/* ieee device used by generic ieee processing code */
 	struct ieee80211_device *ieee;
@@ -2117,7 +2132,7 @@ struct ipw_priv {
 	/* return code for synchronous driver -> daemon commands */
 	s32 daemon_cmd_rc;
 	/* daemon driven work queue */
-	struct work_struct daemon_cmd_work;
+	delayed_work_t daemon_cmd_work;
 	struct workqueue_struct *daemonqueue;
 	struct work_struct daemon_tx_status_sync;
 	wait_queue_head_t wait_daemon_cmd_done;
@@ -2206,11 +2221,12 @@ struct ipw_priv {
 
 	struct ipw_power_mgr power_data;
 
-	struct ieee80211_auth *auth_frame;
-	enum connection_manager_assoc_states auth_state;
+	enum ipw_auth_sequence auth_state;
 
+	struct ipw_frame *assoc_sequence_frame;
 	struct ipw_associate assoc_request;
 	struct ieee80211_network *assoc_network;
+	int association_retries;
 
 	struct ipw_notif_statistics statistics;
 
@@ -2272,33 +2288,34 @@ struct ipw_priv {
 	/* Driver and iwconfig driven work queue */
 	struct workqueue_struct *workqueue;
 
-	struct work_struct up;
-	struct work_struct down;
 	struct work_struct adhoc_check;
 	struct work_struct calibrated_work;
 	struct work_struct associate;
 	struct work_struct disassociate;
 	struct work_struct rx_replenish;
-	struct work_struct request_scan;
-	struct work_struct rf_kill;
-	struct work_struct gather_stats;
 	struct work_struct abort_scan;
 	struct work_struct roam;
-	struct work_struct scan_check;
 	struct work_struct link_up;
 	struct work_struct update_link_led;
 	struct work_struct link_down;
-	struct work_struct associate_timeout;
-	struct work_struct activity_timer;
-	struct work_struct alive_start;
-	struct work_struct resume_work;
 	struct work_struct auth_work;
 	struct work_struct merge_networks;
 	struct work_struct post_associate;
-	struct work_struct rate_scale_flush;
+
+	delayed_work_t up;
+	delayed_work_t down;
+	delayed_work_t activity_timer;
+	delayed_work_t assoc_state_retry;
+	delayed_work_t request_scan;
+	delayed_work_t gather_stats;
+	delayed_work_t scan_check;
+	delayed_work_t rf_kill;
+	delayed_work_t rate_scale_flush;
+	delayed_work_t resume_work;
+	delayed_work_t alive_start;
 
 	/* 802.11h  */
-	struct work_struct report_work;
+	delayed_work_t report_work;
 
 	struct tasklet_struct irq_tasklet;
 
@@ -2548,4 +2565,25 @@ struct ipw_fixed_rate {
 #define IPW_RATE_SCALE_MIN_SUCCESS_TH       8
 #define IPW_RATE_SCALE_DECREASE_TH       1920
 
+#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,14)
+static inline void *kzalloc(size_t size, unsigned __nocast flags)
+{
+	void *ret = kmalloc(size, flags);
+	if (ret)
+		memset(ret, 0, size);
+	return ret;
+}
+#endif
+
+#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,15)
+static inline unsigned compare_ether_addr(const u8 * _a, const u8 * _b)
+{
+	const u16 *a = (const u16 *)_a;
+	const u16 *b = (const u16 *)_b;
+
+	BUILD_BUG_ON(ETH_ALEN != 6);
+	return ((a[0] ^ b[0]) | (a[1] ^ b[1]) | (a[2] ^ b[2])) != 0;
+}
+#endif
+
 #endif				/* __ipw3945_h__ */
